import { error } from "@sveltejs/kit";
import type { CheckResponse, ClientGameData, GameOptions, ServerGameData } from "./types";

type FetchType = typeof fetch;

const metroUrl =
	"https://data.grandlyon.com/geoserver/sytral/ows?SERVICE=WFS&VERSION=2.0.0&request=GetFeature&typename=sytral:tcl_sytral.tcllignemf_2_0_0&outputFormat=application/json&SRSNAME=EPSG:4171&sortBy=gid";
const tramUrl =
	"https://data.grandlyon.com/geoserver/sytral/ows?SERVICE=WFS&VERSION=2.0.0&request=GetFeature&typename=sytral:tcl_sytral.tcllignetram_2_0_0&outputFormat=application/json&SRSNAME=EPSG:4171&sortBy=gid";
const stopsUrl =
	"https://data.grandlyon.com/geoserver/sytral/ows?SERVICE=WFS&VERSION=2.0.0&request=GetFeature&typename=sytral:tcl_sytral.tclarret&outputFormat=application/json&SRSNAME=EPSG:4171&sortBy=gid";

let lazyMetro: [GeoJSON.Feature, string][] | null = null;
let lazyTram: [GeoJSON.Feature, string][] | null = null;
let lazyStops: GeoJSON.Feature[] | null = null;

export async function createGame(
	mode: string,
	stops: GeoJSON.Feature[],
	kv: KVNamespace<string>,
): Promise<ClientGameData> {
	const uuid = crypto.randomUUID();
	const stopNames = stops.map((s) => s.properties!.nom);

	const serverData: ServerGameData = {
		mode,
		stops,
		distances: new Array(stops.length).fill(-1),
		scores: new Array(stops.length).fill(-1),
	};

	await kv.put(`game:${uuid}`, JSON.stringify(serverData), { expirationTtl: 600 });

	return {
		gameId: uuid,
		stopNames,
	};
}

export async function saveScore(
	uuid: string,
	index: number,
	distanceCalc: (latlng: [number, number]) => number,
	scoreCalc: (dist: number) => number,
	kv: KVNamespace<string>,
): Promise<CheckResponse | null> {
	const serverData: ServerGameData | null = await kv.get(`game:${uuid}`, "json");
	if (!serverData) return null;

	if (serverData.scores[index] != -1) {
		return null;
	} else {
		const coords = (serverData.stops[index].geometry as GeoJSON.Point).coordinates;
		const latlng: [number, number] = [coords[1], coords[0]];

		const distance = Math.round(distanceCalc(latlng));
		const score = Math.round(scoreCalc(distance));

		serverData.distances[index] = distance;
		serverData.scores[index] = score;
		await kv.put(`game:${uuid}`, JSON.stringify(serverData), { expirationTtl: 600 });

		return {
			distance,
			score,
			solution: latlng,
		};
	}
}

export async function getMetro(fetch: FetchType): Promise<[GeoJSON.Feature, string][]> {
	if (!lazyMetro) {
		const metro: GeoJSON.FeatureCollection = await fetch(metroUrl).then((r) => r.json());

		lazyMetro = metro.features.map((f) => {
			const components = f.properties!.couleur!.split(" ");
			return [f, `rgb(${components.join(",")})`];
		});
	}

	return lazyMetro;
}

export async function getTram(fetch: FetchType): Promise<[GeoJSON.Feature, string][]> {
	if (!lazyTram) {
		const tram: GeoJSON.FeatureCollection = await fetch(tramUrl).then((r) => r.json());

		lazyTram = tram.features.map((f) => {
			const components = f.properties!.couleur!.split(" ");
			return [f, `rgb(${components.join(",")})`];
		});
	}

	return lazyTram;
}

async function getStops(fetch: FetchType): Promise<GeoJSON.Feature[]> {
	if (!lazyStops) {
		const stops: GeoJSON.FeatureCollection = await fetch(stopsUrl).then((r) => r.json());
		lazyStops = stops.features;
	}

	return lazyStops;
}

export async function getMergedStops(
	fetch: FetchType,
	lineCodes: Set<string>,
): Promise<GeoJSON.Feature[]> {
	const stops = await getStops(fetch);

	const crossingStops = stops.filter((f) => {
		if (f.properties?.desserte) {
			return f.properties.desserte
				.split(",")
				.map((d: string) => d.split(":")[0])
				.some((d: string) => lineCodes.has(d));
		} else {
			return false;
		}
	});

	// merge coordinates of the stops
	const mergedStops = Array.from(new Set(crossingStops.map((f) => f.properties?.nom))).map(
		(nom) => {
			const matching = crossingStops.filter((f) => f.properties?.nom === nom);

			const longs = matching.map((f) => (f.geometry as GeoJSON.Point).coordinates[0]);
			const lonAvg = longs.reduce((a, b) => a + b) / longs.length;

			const lats = matching.map((f) => (f.geometry as GeoJSON.Point).coordinates[1]);
			const latAvg = lats.reduce((a, b) => a + b) / lats.length;

			const desserte = matching.map((f) => f.properties?.desserte).join(",");

			const feature = matching[0];
			feature.geometry = { type: "Point", coordinates: [lonAvg, latAvg] };
			if (feature.properties) feature.properties.desserte = desserte;

			return feature;
		},
	);

	return mergedStops;
}

export function parseOptions(url: URL): GameOptions {
	const mode = url.searchParams.get("mode");
	if (mode !== "easy" && mode !== "hard" && mode !== "extreme demon ultra miguel") {
		error(400, "gamemode is invalid");
	}

	return {
		mode,
		metro: url.searchParams.has("metro"),
		tram: url.searchParams.has("tram"),
	};
}

// https://underscorejs.org/docs/modules/sample.html
export function sample<T>(arr: T[], n: number): T[] {
	const last = arr.length - 1;
	for (let index = 0; index < n; index++) {
		const rand = index + Math.floor(Math.random() * (last - index + 1));
		const temp = arr[index];
		arr[index] = arr[rand];
		arr[rand] = temp;
	}
	return arr.slice(0, n);
}
